@@ -1,10 +1,11 @@
+# pylint:disable=arguments-renamed,isinstance-second-argument-not-valid-type
 try:
     import claripy
 except ImportError:
     claripy = None
 
 from .tagged_object import TaggedObject
-from .utils import get_bits
+from .utils import get_bits, stable_hash, is_none_or_likeable
 
 
 class Expression(TaggedObject):
@@ -27,8 +28,11 @@ class Expression(TaggedObject):
         else:
             return self.likes(atom)
 
+    def __eq__(self, other):
+        return type(self) is type(other) and self.likes(other) and self.idx == other.idx
+
     def likes(self, atom):  # pylint:disable=unused-argument,no-self-use
-        return self == atom
+        raise NotImplementedError()
 
     def replace(self, old_expr, new_expr):
         if self is old_expr:
@@ -82,17 +86,20 @@ class Const(Atom):
         return str(self)
 
     def __str__(self):
-        return "%#x<%d>" % (self.value, self.bits)
+        try:
+            return "%#x<%d>" % (self.value, self.bits)
+        except TypeError:
+            return "%f<%d>" % (self.value, self.bits)
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(self) is type(other) and \
-            self.value == other.value and \
-            self.bits == other.bits
+               self.value == other.value and \
+               self.bits == other.bits
 
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((self.value, self.bits))
+        return stable_hash((self.value, self.bits))
 
     @property
     def sign_bit(self):
@@ -122,7 +129,7 @@ class Tmp(Atom):
     def __str__(self):
         return "t%d" % self.tmp_idx
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(self) is type(other) and \
             self.tmp_idx == other.tmp_idx and \
             self.bits == other.bits
@@ -130,7 +137,7 @@ class Tmp(Atom):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash(('tmp', self.tmp_idx, self.bits))
+        return stable_hash(('tmp', self.tmp_idx, self.bits))
 
     def copy(self) -> 'Tmp':
         return Tmp(self.idx, self.variable, self.tmp_idx, self.bits, **self.tags)
@@ -166,13 +173,10 @@ class Register(Atom):
         else:
             return "%s" % str(self.variable.name)
 
-    def __eq__(self, other):
-        return self.likes(other) and self.idx == other.idx
-
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash(('reg', self.reg_offset, self.bits, self.idx))
+        return stable_hash(('reg', self.reg_offset, self.bits, self.idx))
 
     def copy(self) -> 'Register':
         return Register(self.idx, self.variable, self.reg_offset, self.bits, **self.tags)
@@ -209,7 +213,7 @@ class UnaryOp(Op):
     def __repr__(self):
         return str(self)
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(other) is UnaryOp and \
                self.op == other.op and \
                self.bits == other.bits and \
@@ -218,7 +222,7 @@ class UnaryOp(Op):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((self.op, self.operand, self.bits))
+        return stable_hash((self.op, self.operand, self.bits))
 
     def replace(self, old_expr, new_expr):
         r, replaced_operand = self.operand.replace(old_expr, new_expr)
@@ -265,7 +269,7 @@ class Convert(UnaryOp):
     def __repr__(self):
         return str(self)
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(other) is Convert and \
                self.from_bits == other.from_bits and \
                self.to_bits == other.to_bits and \
@@ -276,10 +280,10 @@ class Convert(UnaryOp):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((self.operand, self.from_bits, self.to_bits, self.bits, self.is_signed))
+        return stable_hash((self.operand, self.from_bits, self.to_bits, self.bits, self.is_signed))
 
     def replace(self, old_expr, new_expr):
-        if self.operand.likes(old_expr):
+        if self.operand == old_expr:
             r = True
             replaced_operand = new_expr
         else:
@@ -294,6 +298,59 @@ class Convert(UnaryOp):
         return Convert(self.idx, self.from_bits, self.to_bits, self.is_signed, self.operand, **self.tags)
 
 
+class Reinterpret(UnaryOp):
+
+    __slots__ = ('from_bits', 'from_type', 'to_bits', 'to_type', )
+
+    def __init__(self, idx, from_bits: int, from_type: str, to_bits: int, to_type: str, operand, **kwargs):
+        super().__init__(idx, 'Reinterpret', operand, **kwargs)
+
+        assert (from_type == "I" and to_type == "F") or (from_type == "F" and to_type == "I")
+
+        self.from_bits = from_bits
+        self.from_type = from_type
+        self.to_bits = to_bits
+        self.to_type = to_type
+
+        self.bits = self.to_bits
+
+    def __str__(self):
+        return f"Reinterpret({self.from_type}{self.from_bits}->{self.to_type}{self.to_bits}, {self.operand})"
+
+    def __repr__(self):
+        return str(self)
+
+    def likes(self, other):
+        return type(other) is Reinterpret and \
+               self.from_bits == other.from_bits and \
+               self.from_type == other.from_type and \
+               self.to_bits == other.to_bits and \
+               self.to_type == other.to_type and \
+               self.operand == other.operand
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash((self.operand, self.from_bits, self.from_type, self.to_bits, self.to_type, ))
+
+    def replace(self, old_expr, new_expr):
+        if self.operand == old_expr:
+            r = True
+            replaced_operand = new_expr
+        else:
+            r, replaced_operand = self.operand.replace(old_expr, new_expr)
+
+        if r:
+            return True, Reinterpret(self.idx, self.from_bits, self.from_type, self.to_bits, self.to_type,
+                                     replaced_operand, **self.tags)
+        else:
+            return False, self
+
+    def copy(self) -> 'Reinterpret':
+        return Reinterpret(self.idx, self.from_bits, self.from_type, self.to_bits, self.to_type, self.operand,
+                           **self.tags)
+
+
 class BinaryOp(Op):
 
     __slots__ = ('operands', 'bits', 'signed', 'variable', 'variable_offset', )
@@ -312,6 +369,7 @@ class BinaryOp(Op):
         'Shl': '<<',
         'Shr': '>>',
         'Sar': '>>a',
+        'CmpF': '==f',
         'CmpEQ': '==',
         'CmpNE': '!=',
         'CmpLT': '<',
@@ -338,7 +396,7 @@ class BinaryOp(Op):
         'CmpGTs': 'CmpLEs',
     }
 
-    def __init__(self, idx, op, operands, signed, variable=None, variable_offset=None, **kwargs):
+    def __init__(self, idx, op, operands, signed, variable=None, variable_offset=None, bits=None, **kwargs):
         depth = max(
             operands[0].depth if isinstance(operands[0], Expression) else 0,
             operands[1].depth if isinstance(operands[1], Expression) else 0,
@@ -347,7 +405,10 @@ class BinaryOp(Op):
 
         assert len(operands) == 2
         self.operands = operands
-        if self.op == 'Cmp':
+
+        if bits is not None:
+            self.bits = bits
+        elif self.op == 'CmpF':
             self.bits = 32  # floating point comparison
         elif self.op.startswith("Cmp"):
             self.bits = 1
@@ -369,17 +430,17 @@ class BinaryOp(Op):
     def __repr__(self):
         return "%s(%s, %s)" % (self.verbose_op, self.operands[0], self.operands[1])
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(other) is BinaryOp and \
                self.op == other.op and \
                self.bits == other.bits and \
                self.signed == other.signed and \
-               self.operands == other.operands
+               is_none_or_likeable(self.operands, other.operands, is_list=True)
 
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((self.op, tuple(self.operands), self.bits, self.signed))
+        return stable_hash((self.op, tuple(self.operands), self.bits, self.signed))
 
     def has_atom(self, atom, identity=True):
         if super().has_atom(atom, identity=identity):
@@ -388,9 +449,9 @@ class BinaryOp(Op):
         for op in self.operands:
             if identity and op == atom:
                 return True
-            if not identity and isinstance(op, Atom) and op.likes(atom):
+            if not identity and isinstance(op, Expression) and op.likes(atom):
                 return True
-            if isinstance(op, Atom) and op.has_atom(atom, identity=identity):
+            if isinstance(op, Expression) and op.has_atom(atom, identity=identity):
                 return True
         return False
 
@@ -413,6 +474,7 @@ class BinaryOp(Op):
 
         if r0 or r1:
             return True, BinaryOp(self.idx, self.op, [ replaced_operand_0, replaced_operand_1 ], self.signed,
+                                  bits=self.bits,
                                   **self.tags)
         else:
             return False, self
@@ -430,7 +492,7 @@ class BinaryOp(Op):
 
     def copy(self) -> 'BinaryOp':
         return BinaryOp(self.idx, self.op, self.operands[::], self.signed, variable=self.variable,
-                        variable_offset=self.variable_offset, **self.tags)
+                        variable_offset=self.variable_offset, bits=self.bits, **self.tags)
 
 
 class Load(Expression):
@@ -468,16 +530,26 @@ class Load(Expression):
         return self.addr.has_atom(atom, identity=identity)
 
     def replace(self, old_expr, new_expr):
-        r, replaced_addr = self.addr.replace(old_expr, new_expr)
+        if self.addr == old_expr:
+            r = True
+            replaced_addr = new_expr
+        else:
+            r, replaced_addr = self.addr.replace(old_expr, new_expr)
 
         if r:
             return True, Load(self.idx, replaced_addr, self.size, self.endness, **self.tags)
         else:
             return False, self
 
-    def __eq__(self, other):
+    def _likes_addr(self, other_addr):
+        if hasattr(self.addr, "likes") and hasattr(other_addr, "likes"):
+            return self.addr.likes(other_addr)
+
+        return self.addr == other_addr
+
+    def likes(self, other):
         return type(other) is Load and \
-               self.addr == other.addr and \
+               self._likes_addr(other.addr) and \
                self.size == other.size and \
                self.endness == other.endness and \
                self.guard == other.guard and \
@@ -486,7 +558,7 @@ class Load(Expression):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash(('Load', self.addr, self.size, self.endness))
+        return stable_hash(('Load', self.addr, self.size, self.endness))
 
     def copy(self) -> 'Load':
         return Load(self.idx, self.addr, self.size, self.endness, variable=self.variable,
@@ -517,8 +589,17 @@ class ITE(Expression):
     def __str__(self):
         return "((%s) ? (%s) : (%s))" % (self.cond, self.iftrue, self.iffalse)
 
+    def likes(self, atom):
+        return type(atom) is ITE and \
+            self.cond == atom.cond and \
+            self.iffalse == atom.iffalse and \
+            self.iftrue == atom.iftrue and \
+            self.bits == atom.bits
+
+    __hash__ = TaggedObject.__hash__
+
     def _hash_core(self):
-        return hash((ITE, self.cond, self.iffalse, self.iftrue, self.bits))
+        return stable_hash((ITE, self.cond, self.iffalse, self.iftrue, self.bits))
 
     def has_atom(self, atom, identity=True):
         if super().has_atom(atom, identity=identity):
@@ -549,22 +630,20 @@ class ITE(Expression):
 
 class DirtyExpression(Expression):
 
-    __slots__ = ('dirty_expr', )
+    __slots__ = ('dirty_expr', 'bits', )
 
-    def __init__(self, idx, dirty_expr, **kwargs):
+    def __init__(self, idx, dirty_expr, bits=None, **kwargs):
         super().__init__(idx, 1, **kwargs)
         self.dirty_expr = dirty_expr
+        self.bits = bits
 
-    def replace(self, old_expr, new_expr):
-        return False, self
-
-    def __eq__(self, other):
+    def likes(self, other):
         return type(other) is DirtyExpression and other.dirty_expr == self.dirty_expr
 
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((DirtyExpression, self.dirty_expr))
+        return stable_hash((DirtyExpression, self.dirty_expr))
 
     def __repr__(self):
         return "DirtyExpression (%s)" % type(self.dirty_expr)
@@ -573,8 +652,78 @@ class DirtyExpression(Expression):
         return "[D] %s" % str(self.dirty_expr)
 
     def copy(self) -> 'DirtyExpression':
-        return DirtyExpression(self.idx, self.dirty_expr, **self.tags)
+        return DirtyExpression(self.idx, self.dirty_expr, bits=self.bits, **self.tags)
 
+    def replace(self, old_expr, new_expr):
+        if old_expr is self.dirty_expr:
+            return True, DirtyExpression(self.idx, new_expr, bits=self.bits, **self.tags)
+
+        replaced, new_dirty_expr = self.dirty_expr.replace(old_expr, new_expr)
+        if replaced:
+            return True, DirtyExpression(self.idx, new_dirty_expr, bits=self.bits, **self.tags)
+        else:
+            return False, self
+
+    @property
+    def size(self):
+        return self.bits // 8
+
+
+class VEXCCallExpression(Expression):
+
+    __slots__ = ('cee_name', 'operands', 'bits', )
+
+    def __init__(self, idx, cee_name, operands, bits=None, **kwargs):
+        super().__init__(idx, max(operand.depth for operand in operands), **kwargs)
+        self.cee_name = cee_name
+        self.operands = operands
+        self.bits = bits
+
+    def likes(self, other):
+        return type(other) is VEXCCallExpression and \
+               other.cee_name == self.cee_name and \
+               len(self.operands) == len(other.operands) and \
+               self.bits == other.bits and \
+               all(op1.likes(op2) for op1, op2 in zip(other.operands, self.operands))
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash((VEXCCallExpression, self.cee_name, self.bits, tuple(self.operands)))
+
+    def __repr__(self):
+        return f"VEXCCallExpression [{self.cee_name}]"
+
+    def __str__(self):
+        operands_str = ", ".join(repr(op) for op in self.operands)
+        return f"{self.cee_name}({operands_str})"
+
+    def copy(self) -> 'VEXCCallExpression':
+        return VEXCCallExpression(self.idx, self.cee_name, self.operands, bits=self.bits, **self.tags)
+
+    def replace(self, old_expr, new_expr):
+        new_operands = [ ]
+        replaced = False
+        for operand in self.operands:
+            if operand is old_expr:
+                new_operands.append(new_expr)
+                replaced = True
+            else:
+                operand_replaced, new_operand = operand.replace(old_expr, new_expr)
+                if operand_replaced:
+                    new_operands.append(new_operand)
+                    replaced = True
+                else:
+                    new_operands.append(operand)
+
+        if replaced:
+            return True, VEXCCallExpression(self.idx, self.cee_name, tuple(new_operands), bits=self.bits, **self.tags)
+        else:
+            return False, self
+
+    @property
+    def size(self):
+        return self.bits // 8
 
 #
 # Special (Dummy) expressions
@@ -611,7 +760,7 @@ class BasePointerOffset(Expression):
             return "%s%+d" % (self.base, self.offset)
         return "%s+%s" % (self.base, self.offset)
 
-    def __eq__(self, other):
+    def likes(self, other):
         return type(other) is type(self) and \
                self.bits == other.bits and \
                self.base == other.base and \
@@ -620,7 +769,7 @@ class BasePointerOffset(Expression):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return hash((self.bits, self.base, self.offset))
+        return stable_hash((self.bits, self.base, self.offset))
 
     def replace(self, old_expr, new_expr):
         if isinstance(self.base, Expression):
